@@ -1,7 +1,7 @@
 /*
  * fields.c
  *
- * Copyright (c) Chris Putnam 2003-2013
+ * Copyright (c) Chris Putnam 2003-2018
  *
  * Source code released under the GPL version 2
  *
@@ -36,8 +36,8 @@ fields_free( fields *f )
 	int i;
 
 	for ( i=0; i<f->max; ++i ) {
-		newstr_free( &(f->tag[i]) );
-		newstr_free( &(f->data[i]) );
+		str_free( &(f->tag[i]) );
+		str_free( &(f->data[i]) );
 	}
 	if ( f->tag )   free( f->tag );
 	if ( f->data )  free( f->data );
@@ -47,13 +47,20 @@ fields_free( fields *f )
 	fields_init( f );
 }
 
+void
+fields_delete( fields *f )
+{
+	fields_free( f );
+	free( f );
+}
+
 static int
 fields_alloc( fields *f )
 {
 	int i, alloc = 20;
 
-	f->tag   = (newstr *) malloc( sizeof(newstr) * alloc );
-	f->data  = (newstr *) malloc( sizeof(newstr) * alloc );
+	f->tag   = (str *) malloc( sizeof(str) * alloc );
+	f->data  = (str *) malloc( sizeof(str) * alloc );
 	f->used  = (int *)    calloc( alloc, sizeof(int) );
 	f->level = (int *)    calloc( alloc, sizeof(int) );
 	if ( !f->tag || !f->data || !f->used || !f->level ){
@@ -68,8 +75,8 @@ fields_alloc( fields *f )
 	f->max = alloc;
 	f->n = 0;
 	for ( i=0; i<alloc; ++i ) {
-		newstr_init( &(f->tag[i]) );
-		newstr_init( &(f->data[i]) );
+		str_init( &(f->tag[i]) );
+		str_init( &(f->data[i]) );
 	}
 	return FIELDS_OK;
 }
@@ -77,12 +84,12 @@ fields_alloc( fields *f )
 static int
 fields_realloc( fields *f )
 {
-	newstr *newtags, *newdata;
+	str *newtags, *newdata;
 	int *newused, *newlevel;
 	int i, alloc = f->max * 2;
 
-	newtags = (newstr*) realloc( f->tag, sizeof(newstr) * alloc );
-	newdata = (newstr*) realloc( f->data, sizeof(newstr) * alloc );
+	newtags = (str*) realloc( f->tag, sizeof(str) * alloc );
+	newdata = (str*) realloc( f->data, sizeof(str) * alloc );
 	newused = (int*)    realloc( f->used, sizeof(int) * alloc );
 	newlevel= (int*)    realloc( f->level, sizeof(int) * alloc );
 
@@ -97,17 +104,18 @@ fields_realloc( fields *f )
 	f->max = alloc;
 
 	for ( i=f->n; i<alloc; ++i ) {
-		newstr_init( &(f->tag[i]) );
-		newstr_init( &(f->data[i]) );
+		str_init( &(f->tag[i]) );
+		str_init( &(f->data[i]) );
 	}
 
 	return FIELDS_OK;
 }
 
 int
-fields_add( fields *f, char *tag, char *data, int level )
+_fields_add( fields *f, char *tag, char *data, int level, int mode )
 {
-	int i, status, found;
+	int i, n, status;
+	str *t, *d;
 
 	if ( !tag || !data ) return FIELDS_OK;
 
@@ -119,38 +127,44 @@ fields_add( fields *f, char *tag, char *data, int level )
 		if ( status!=FIELDS_OK ) return status;
 	}
 
-	/* Don't duplicate identical entries */
-	found = 0;
-	for ( i=0; i<f->n && !found; ++i ) {
-		if ( f->level[i]==level &&
-		     !strcasecmp( f->tag[i].data, tag ) &&
-		     !strcasecmp( f->data[i].data, data ) ) found = 1;
-	}
-	if ( !found ) {
-		f->used[ f->n ]  = 0;
-		f->level[ f->n ] = level;
-		newstr_strcpy( &(f->tag[f->n]), tag );
-		newstr_strcpy( &(f->data[f->n]), data );
-		if ( newstr_memerr( &(f->tag[f->n]) ) ||
-		     newstr_memerr( &(f->data[f->n] ) ) )
-			return FIELDS_ERR;
-		f->n++;
+	/* Don't duplicate identical entries if FIELDS_NO_DUPS */
+	if ( mode == FIELDS_NO_DUPS ) {
+		for ( i=0; i<f->n; i++ ) {
+			t = &(f->tag[i]);
+			d = &(f->data[i]);
+			if ( f->level[i]==level &&
+			     !strcasecmp( str_cstr( t ), tag ) &&
+			     !strcasecmp( str_cstr( d ), data ) )
+				return FIELDS_OK;
+		}
 	}
+
+	n = f->n;
+	f->used[ n ]  = 0;
+	f->level[ n ] = level;
+	str_strcpyc( &(f->tag[n]), tag );
+	str_strcpyc( &(f->data[n]), data );
+
+	if ( str_memerr( &(f->tag[n]) ) || str_memerr( &(f->data[n] ) ) )
+		return FIELDS_ERR;
+
+	f->n++;
+
 	return FIELDS_OK;
 }
 
 int
-fields_add_tagsuffix( fields *f, char *tag, char *suffix,
-		char *data, int level )
+_fields_add_tagsuffix( fields *f, char *tag, char *suffix,
+		char *data, int level, int mode )
 {
-	newstr newtag;
+	str newtag;
 	int ret;
 
-	newstr_init( &newtag );
-	newstr_mergestrs( &newtag, tag, suffix, NULL );
-	if ( newstr_memerr( &newtag ) ) ret = FIELDS_ERR;
-	else ret = fields_add( f, newtag.data, data, level );
-	newstr_free( &newtag );
+	str_init( &newtag );
+	str_mergestrs( &newtag, tag, suffix, NULL );
+	if ( str_memerr( &newtag ) ) ret = FIELDS_ERR;
+	else ret = _fields_add( f, newtag.data, data, level, mode );
+	str_free( &newtag );
 
 	return ret;
 }
@@ -161,7 +175,7 @@ fields_add_tagsuffix( fields *f, char *tag, char *suffix,
  *
  * level==LEVEL_ANY is a special flag meaning any level can match
  */
-inline int
+int
 fields_match_level( fields *f, int n, int level )
 {
 	if ( level==LEVEL_ANY ) return 1;
@@ -174,28 +188,28 @@ fields_match_level( fields *f, int n, int level )
  * returns 1 if tag matches, 0 if not
  *
  */
-inline int
+int
 fields_match_tag( fields *info, int n, char *tag )
 {
 	if ( !strcmp( fields_tag( info, n, FIELDS_CHRP ), tag ) ) return 1;
 	return 0;
 }
 
-inline int
+int
 fields_match_casetag( fields *info, int n, char *tag )
 {
 	if ( !strcasecmp( fields_tag( info, n, FIELDS_CHRP ), tag ) ) return 1;
 	return 0;
 }
 
-inline int
+int
 fields_match_tag_level( fields *info, int n, char *tag, int level )
 {
 	if ( !fields_match_level( info, n, level ) ) return 0;
 	return fields_match_tag( info, n, tag );
 }
 
-inline int
+int
 fields_match_casetag_level( fields *info, int n, char *tag, int level )
 {
 	if ( !fields_match_level( info, n, level ) ) return 0;
@@ -205,7 +219,7 @@ fields_match_casetag_level( fields *info, int n, char *tag, int level )
 /* fields_find()
  *
  * Return position [0,f->n) for match of the tag.
- * Return -1 if tag isn't found.
+ * Return FIELDS_NOTFOUND if tag isn't found.
  */
 int
 fields_find( fields *f, char *tag, int level )
@@ -223,7 +237,7 @@ fields_find( fields *f, char *tag, int level )
 		}
 	}
 
-	return -1;
+	return FIELDS_NOTFOUND;
 }
 
 int
@@ -266,37 +280,46 @@ int
 fields_replace_or_add( fields *f, char *tag, char *data, int level )
 {
 	int n = fields_find( f, tag, level );
-	if ( n==-1 ) return fields_add( f, tag, data, level );
+	if ( n==FIELDS_NOTFOUND ) return fields_add( f, tag, data, level );
 	else {
-		newstr_strcpy( &(f->data[n]), data );
-		if ( newstr_memerr( &(f->data[n]) ) ) return FIELDS_ERR;
+		str_strcpyc( &(f->data[n]), data );
+		if ( str_memerr( &(f->data[n]) ) ) return FIELDS_ERR;
 		return FIELDS_OK;
 	}
 }
 
 char *fields_null_value = "\0";
 
-inline int
+int
 fields_used( fields *f, int n )
 {
 	if ( n >= 0 && n < f->n ) return f->used[n];
 	else return 0;
 }
 
-inline int
+int
+fields_notag( fields *f, int n )
+{
+	str *t;
+	if ( n >= 0 && n < f->n ) {
+		t = &( f->tag[n] );
+		if ( t->len > 0 ) return 0;
+	}
+	return 1;
+}
+
+int
 fields_nodata( fields *f, int n )
 {
-	newstr *d;
+	str *d;
 	if ( n >= 0 && n < f->n ) {
 		d = &( f->data[n] );
 		if ( d->len > 0 ) return 0;
-		return 1;
-	} else {
-		return 1;
 	}
+	return 1;
 }
 
-inline int
+int
 fields_num( fields *f )
 {
 	return f->n;
@@ -311,7 +334,7 @@ fields_num( fields *f )
  * If the length of the tagged value is zero and the mode is
  * FIELDS_STRP_NOLEN or FIELDS_CHRP_NOLEN, return a pointer to
  * a static null string as the data field could be new due to
- * the way newstr handles initialized strings with no data.
+ * the way str handles initialized strings with no data.
  *
  */
 
@@ -368,10 +391,10 @@ fields_level( fields *f, int n )
 void *
 fields_findv( fields *f, int level, int mode, char *tag )
 {
-	int i, found = -1;
+	int i, found = FIELDS_NOTFOUND;
 	intptr_t retn;
 
-	for ( i=0; i<f->n && found==-1; ++i ) {
+	for ( i=0; i<f->n && found==FIELDS_NOTFOUND; ++i ) {
 
 		if ( !fields_match_level( f, i, level ) ) continue;
 		if ( !fields_match_casetag( f, i, tag ) ) continue;
@@ -386,7 +409,7 @@ fields_findv( fields *f, int level, int mode, char *tag )
 		}
 	}
 
-	if ( found==-1 ) return NULL;
+	if ( found==FIELDS_NOTFOUND ) return NULL;
 
 	if ( mode & FIELDS_SETUSE_FLAG )
 		fields_setused( f, found );
@@ -419,30 +442,35 @@ fields_findv_firstof( fields *f, int level, int mode, ... )
 	return NULL;
 }
 
-static void
+static int
 fields_findv_each_add( fields *f, int mode, int n, vplist *a )
 {
-	intptr_t retn;
+	int status;
+	void *v;
 
-	if ( n<0 || n>= f->n ) return;
+	if ( n<0 || n>= f->n ) return FIELDS_OK;
 
 	if ( mode & FIELDS_SETUSE_FLAG )
 		fields_setused( f, n );
 
 	if ( mode & FIELDS_STRP_FLAG ) {
-		vplist_add( a, (void *) &(f->data[n]) );
+		v = ( void * ) &( f->data[n] );
 	} else if ( mode & FIELDS_POSP_FLAG ) {
-		retn = n;
-		vplist_add( a, (void *) retn );
+		v = ( void * )( (long) n );
 	} else {
-		vplist_add( a, (void *) f->data[n].data );
+		v = ( void * ) str_cstr( &( f->data[n] ) );
 	}
+
+	status = vplist_add( a, v );
+
+	if ( status==VPLIST_OK ) return FIELDS_OK;
+	else return FIELDS_ERR;
 }
 
-void
+int
 fields_findv_each( fields *f, int level, int mode, vplist *a, char *tag )
 {
-	int i;
+	int i, status;
 
 	for ( i=0; i<f->n; ++i ) {
 
@@ -450,57 +478,95 @@ fields_findv_each( fields *f, int level, int mode, vplist *a, char *tag )
 		if ( !fields_match_casetag( f, i, tag ) ) continue;
 
 		if ( f->data[i].len!=0 ) {
-			fields_findv_each_add( f, mode, i, a );
+			status = fields_findv_each_add( f, mode, i, a );
+			if ( status!=FIELDS_OK ) return status;
 		} else {
 			if ( mode & FIELDS_NOLENOK_FLAG ) {
-				fields_findv_each_add( f, mode, i, a );
+				status = fields_findv_each_add( f, mode, i, a );
+				if ( status!=FIELDS_OK ) return status;
 			} else {
 				f->used[i] = 1; /* Suppress "noise" of unused */
 			}
 		}
 
 	}
+
+	return FIELDS_OK;
 }
 
-void
+static int
+fields_build_tags( va_list argp, vplist *tags )
+{
+	int status;
+	char *tag;
+
+	while ( ( tag = ( char * ) va_arg( argp, char * ) ) ) {
+		status = vplist_add( tags, tag );
+		if ( status!=VPLIST_OK ) return FIELDS_ERR;
+	}
+
+	return FIELDS_OK;
+}
+
+static int
+fields_match_casetags( fields *f, int n, vplist *tags )
+{
+	int i;
+
+	for ( i=0; i<tags->n; ++i )
+		if ( fields_match_casetag( f, n, vplist_get( tags, i ) ) ) return 1;
+
+	return 0;
+}
+
+int
 fields_findv_eachof( fields *f, int level, int mode, vplist *a, ... )
 {
+	int i, status;
 	va_list argp;
 	vplist tags;
-	char *tag;
-	int i, j, found;
 
 	vplist_init( &tags );
 
 	/* build list of tags to search for */
 	va_start( argp, a );
-	while ( ( tag = ( char * ) va_arg( argp, char * ) ) )
-		vplist_add( &tags, tag );
+	status = fields_build_tags( argp, &tags );
 	va_end( argp );
+	if ( status!=FIELDS_OK ) goto out;
 
 	/* search list */
 	for ( i=0; i<f->n; ++i ) {
 
 		if ( !fields_match_level( f, i, level ) ) continue;
+		if ( !fields_match_casetags( f, i, &tags ) ) continue;
 
-		found = 0;
-		for ( j=0; j<tags.n && !found; ++j ) {
-			if ( fields_match_casetag( f, i, vplist_get( &tags, j ) ) )
-				found = 1;
-		}
-		if ( !found ) continue;
-
-		if ( f->data[i].len!=0 ) {
-			fields_findv_each_add( f, mode, i, a );
+		if ( f->data[i].len!=0 || ( mode & FIELDS_NOLENOK_FLAG ) ) {
+			status = fields_findv_each_add( f, mode, i, a );
+			if ( status!=FIELDS_OK ) goto out;
 		} else {
-			if ( mode & FIELDS_NOLENOK_FLAG ) {
-				fields_findv_each_add( f, mode, i, a );
-			} else {
-				f->used[i] = 1; /* Suppress "noise" of unused */
-			}
+			f->used[i] = 1; /* Suppress "noise" of unused */
 		}
 
 	}
 
+out:
 	vplist_free( &tags );
+	return status;
 }
+
+void
+fields_report( fields *f, FILE *fp )
+{
+	int i, n;
+	n = fields_num( f );
+	fprintf( fp, "# NUM   level = LEVEL   'TAG' = 'VALUE'\n" );
+	for ( i=0; i<n; ++i ) {
+		fprintf( stderr, "%d\tlevel = %d\t'%s' = '%s'\n",
+			i+1,
+			fields_level( f, i ),
+			(char*)fields_tag( f, i, FIELDS_CHRP_NOUSE ),
+			(char*)fields_value( f, i, FIELDS_CHRP_NOUSE )
+		);
+	}
+}
+
