@@ -1,14 +1,14 @@
 /*
  * unicode.h
  *
- * Copyright (c) Chris Putnam 2004-2013
+ * Copyright (c) Chris Putnam 2004-2018
  *
  * Program and source code released under the GPL version 2
  */
 #ifndef UNICODE_H
 #define UNICODE_H
 
-#include "newstr.h"
+#include "str.h"
 
 #define UNICODE_SYMBOL    (1)
 #define UNICODE_UPPER     (2) /* Uppercase letter */
@@ -17,6 +17,6 @@
 #define UNICODE_MIXEDCASE ( UNICODE_UPPER | UNICODE_LOWER )
 
 extern unsigned short unicode_utf8_classify( char *p );
-extern unsigned short unicode_utf8_classify_newstr( newstr *s );
+extern unsigned short unicode_utf8_classify_str( str *s );
 
 #endif
